import copy 
import logging
import json
import copy
import cv2

from foglamp.common import logger
import filter_ingest

import numpy as np

__author__ = "Bryce Kroencke"
__copyright__ = "Copyright (c) 2023 Dianomic Systems"
__license__ = "Apache 2.0"
__version__ = "${VERSION}"

_LOGGER = logger.setup(__name__, level=logging.DEBUG)

PLUGIN_NAME = "image-bounding-box"

_DEFAULT_CONFIG = {
    'plugin': {
        'description': 'Crop out or draw bounding boxes on an image',
        'type': 'string',
        'default': PLUGIN_NAME,
        'readonly': 'true'
    },
    'enable': {
        'description': 'Enable plugin',
        'type': 'boolean',
        'default': 'true',
        'displayName': 'Enabled',
        'order': "1"
    },
    'datapoint': {
        'description': 'Data Point containing image data to write on',
        'type': 'string',
        'default': 'img',
        'displayName': 'Image Data Point',
        'order': "2"
    },
    'forward_original': {
        'description': 'If false, the original image will be stripped from the readings leaving only the bounding box image/images',
        'type': 'boolean',
        'default': 'true',
        'displayName': 'Forward Original Image',
        'order': "3"
    },
    'box_datapoint': {
        'description': 'Data point containing the bounding boxes from an ML Inference',
        'type': 'string',
        'default': 'boxes',
        'displayName': 'Bounding Box Data Point',
        'order': "4"
    },
    'box_operation': {
        'description': 'Specifies if the bounding boxes should be cropped or drawn on to the original image',
        'type': 'enumeration',
        'default': 'Crop Bounding Boxes',
        'options': ['Crop Bounding Boxes', 'Draw Bounding Boxes'],
        'order': '5',
        'displayName': 'Bounding Box Operation'
    },
    'crop_prefix': {
        'description': 'Prefix used for cropped images',
        'type': 'string',
        'default': 'cropped_',
        'displayName': 'Crop Prefix',
        'order': '6'
    },
    'topGrowPct': {
        'description': 'Percentage to grow/shrink the top of the box',
        'type': 'float',
        'default': '0.0',
        'displayName': 'Top Grow %',
        'order': '7'
    },
    'bottomGrowPct': {
        'description': 'Percentage to grow/shrink the bottom of the box',
        'type': 'float',
        'default': '0.0',
        'displayName': 'Bottom Grow %',
        'order': '8'
    },
    'leftGrowPct': {
        'description': 'Percentage to grow/shrink the left of the box',
        'type': 'float',
        'default': '0.0',
        'displayName': 'Left Grow %',
        'order': '9'
    },
    'rightGrowPct': {
        'description': 'Percentage to grow/shrink the right of the box',
        'type': 'float',
        'default': '0.0',
        'displayName': 'Right Grow %',
        'order': '10'
    },
    'max_boxes': {
        'description': 'Maximum number of boxes to crop or draw. Leave empty for no limit.',
        'type': 'integer',
        'default': '0',
        'displayName': 'Max Boxes',
        'order': '11'
    }
}

def plugin_info():
    return {
        'name': PLUGIN_NAME,
        'version': '3.0.0-4-g483b3c6',
        'mode': "none",
        'type': 'filter',
        'interface': '1.0',
        'config': _DEFAULT_CONFIG
    }

def plugin_init(config, ingest_ref, callback):
    data = copy.deepcopy(config)
    data['callback'] = callback
    data['ingest_ref'] = ingest_ref
    return data

def plugin_reconfigure(handle, new_config):
    _LOGGER.info("Old config {} \n new config {} for {} plugin ".format(handle, new_config, PLUGIN_NAME))
    new_handle = copy.deepcopy(new_config)
    new_handle['callback'] = handle['callback']
    new_handle['ingest_ref'] = handle['ingest_ref']
    return new_handle

def plugin_shutdown(handle):
    handle['callback'] = None
    handle['ingest_ref'] = None
    _LOGGER.info('{} filter plugin shutdown.'.format(PLUGIN_NAME))

def adjust_bounding_box(img, box, growth):
    h, w = img.shape[:2]
    ymin = int(max(0, box[0] * h))
    xmin = int(max(0, box[1] * w))
    ymax = int(min(h, box[2] * h))
    xmax = int(min(w, box[3] * w))

    cy = (ymin + ymax) / 2
    cx = (xmin + xmax) / 2

    top_dist = cy - ymin
    bottom_dist = ymax - cy
    left_dist = cx - xmin
    right_dist = xmax - cx

    new_ymin = int(max(0, cy - top_dist * (1 + growth['top'])))
    new_ymax = int(min(h, cy + bottom_dist * (1 + growth['bottom'])))
    new_xmin = int(max(0, cx - left_dist * (1 + growth['left'])))
    new_xmax = int(min(w, cx + right_dist * (1 + growth['right'])))

    return new_ymin, new_xmin, new_ymax, new_xmax

def crop_box(img, box, mode, growth):
    ymin, xmin, ymax, xmax = adjust_bounding_box(img, box, growth)
    if mode == 'Crop Bounding Boxes':
        return img[ymin:ymax, xmin:xmax, :].astype(np.uint8)
    else:
        return cv2.rectangle(img, (xmin, ymin), (xmax, ymax), (10, 255, 0), 2)

def plugin_ingest(handle, data):
    if handle['enable']['value'] == 'false':
        filter_ingest.filter_ingest_callback(handle['callback'], handle['ingest_ref'], data)
        return

    mode = handle['box_operation']['value']
    prefix = handle['crop_prefix']['value']
    growth = {
        'top': float(handle['topGrowPct']['value']),
        'bottom': float(handle['bottomGrowPct']['value']),
        'left': float(handle['leftGrowPct']['value']),
        'right': float(handle['rightGrowPct']['value'])
    }
    max_boxes = int(handle['max_boxes']['value'])


    for elem in data:
        boxes = json.loads(elem['readings'][handle['box_datapoint']['value']])
        if handle['datapoint']['value'] not in elem['readings'].keys():
            break
        img = elem['readings'][handle['datapoint']['value']].copy()
        if handle['forward_original']['value'] == 'false':
            del elem['readings'][handle['datapoint']['value']]
        
        if isinstance(boxes, list):
            if all(isinstance(box, list) for box in boxes):
                for i, box in enumerate(boxes):
                    if max_boxes is not 0 and i >= max_boxes:
                        break
                    result = crop_box(img, box, mode, growth)
                    if mode == 'Crop Bounding Boxes':
                        elem['readings'][f"{prefix}{i}"] = result
                    else:
                        img = result
                if mode == 'Draw Bounding Boxes':
                    elem['readings'][handle['datapoint']['value'] + '_boundingbox'] = img
            else:
                result = crop_box(img, boxes, mode, growth)
                if mode == 'Crop Bounding Boxes':
                    elem['readings'][f"{prefix}0"] = result
                else:
                    elem['readings'][handle['datapoint']['value'] + '_boundingbox'] = result

    filter_ingest.filter_ingest_callback(handle['callback'], handle['ingest_ref'], data)
    _LOGGER.debug("{} filter ingest done".format(PLUGIN_NAME))